#!/usr/bin/env python
#
# Copyright (C) 2014 - Renzo Davoli
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# 

import sys, copy, json, argparse, os

inch=2.54
#inch=1

class mille:
	'''
	pcb object: size in mm, width/height in .1" elements.
	table: the same integer represents the same track
	drill: boolean (in the future values can represent different diameters)
	rear: for two sided pcb, not yet implemented
	'''
	def __init__(self):
		self.size=(0,0)
		self.width,self.height=0,0
		self.table=[]
		self.rear=[]
		self.drill=[]
	def valid(self,i,j):
		return True if i>0 and j>0 and i<self.width+1 and j<self.height+1 else False
	def new(self,size=(120,100),twosided=0):
		self.width,self.height=size
		self.size=size
		self.width,self.height=int(self.width//inch),int(self.height//inch)
		self.drill=[[0 for _ in range(self.width+2)] for _ in range(self.height+2)]
		self.table=[[1 if self.valid(i,j) else 0 for i in range(self.width+2)]
			for j in range(self.height+2)]
		if twosided:
			self.rear=[[1 if self.valid(i,j) else 0 for i in range(self.width+2)]
				for j in range(self.height+2)]
	def reversei(self):
		for i in range((self.height+2)//2):
			self.table[i],self.table[self.height+1-i]=self.table[self.height+1-i],self.table[i]
			self.drill[i],self.drill[self.height+1-i]=self.drill[self.height+1-i],self.drill[i]
			if self.rear:
				self.rear[i],self.rear[self.height+1-i]=self.rear[self.height+1-i],self.rear[i]
	def reversej(self):
		for i in range(self.height+2):
			for j in range((self.width+2)//2):
				self.table[i][j],self.table[i][self.width+1-j]=self.table[i][self.width+1-j],self.table[i][j]
				self.drill[i][j],self.drill[i][self.width+1-j]=self.drill[i][self.width+1-j],self.drill[i][j]
				if self.rear:
					self.rear[i][j],self.rear[i][self.width+1-j]=self.rear[i][self.width+1-j],self.rear[i][j]

xoff,yoff=0,0

class defargs:
	uplevel=2.0
	drilllevel=-2.2
	zspeed=50
	g0speed=2000
	g1speed=100

def loadrc():
	try:
		with open(os.path.expanduser("~/.milleforirc"), "r") as f:
			loaded=json.load(f)
			for f in loaded:
				defargs.__dict__[f]=loaded[f]
	except IOError:
		pass

def main():
	pcb=mille()
	loadrc()
	parser = argparse.ArgumentParser()
	parser.add_argument("input_file", help="millefori input file")
	parser.add_argument("-u", "--uplevel", type=float, help="traveling z-level ({})".format(defargs.uplevel), default=defargs.uplevel)
	parser.add_argument("-d", "--drilllevel", type=float, help="drill down to this z-level ({})".format(defargs.drilllevel), default=defargs.drilllevel)
	parser.add_argument("-z", "--zspeed", type=int, help="drilling z speed ({})".format(defargs.zspeed), default=defargs.zspeed)
	parser.add_argument("--g0speed", type=int, help="moving speed ({})".format(defargs.g0speed), default=defargs.g0speed)
	parser.add_argument("--g1speed", type=int, help="working speed ({})".format(defargs.g1speed), default=defargs.g1speed)
	parser.add_argument("-v", "--verbose", action="store_true", help="verbose output (add comments)")
	parser.add_argument("-c", '--centered', action="store_true", help='xy centered')

	args = parser.parse_args()
	args.speed=[args.g0speed,args.g1speed]
	pcb.filename=args.input_file
	if file_load(pcb) == False:
		print "Error opening file", filename
	sx,sy=pcb.size
	if args.centered:
		args.xoff=max((sx-pcb.width*inch)/2,0)
		args.yoff=max((sy-pcb.height*inch)/2,0)
	else:
		args.xoff=args.yoff=0
	pcb.reversei()
	pcb.reversej()
	computedrill(pcb,args,0,0)

# GCODE COMMANDS
def up(args):
	print "G1 Z{} F{}".format(args.uplevel,args.zspeed)

def down(args):
	print "G1 Z{} F{}".format(args.downlevel,args.zspeed)

def downdrill(args):
	print "G1 Z{} F{}".format(args.drilllevel,args.zspeed)

def ggd(args,g,x,y):
	print "G{} X{} Y{} F{}".format(g,x*inch-inch/2+args.xoff,y*inch-inch/2+args.yoff,args.speed[g])

def gg(args,g,x,y):
	print "G{} X{} Y{} F{}".format(g,x*inch+args.xoff,y*inch+args.yoff,args.speed[g])

'''
find the nearest hole to drill
'''
def nearest(pcb,x,y):
	minp=(0,0)
	mind=(pcb.width ** 2 + pcb.height ** 2) ** 0.5
	for i in range(1,pcb.width+1):
		for j in range(1,pcb.height+1):
			if pcb.drill[j][i] != 0:
				dx=x-i;
				dy=y-j;
				d=(dx*dx + dy*dy) ** 0.5
#				d=abs(dx)+abs(dy)
				if d<mind:
					minp=(i,j)
					mind=d
	return minp
	
'''
generate the gcode to drill holes in a pcb:
while there are holes to drill:
	find the nearest hole
	drill it
	tag it as done
'''
def computedrill(pcb,args,x,y):
	up(args)
	while True:
		x,y=nearest(pcb,x,y)
		if x == 0:
			break
		ggd(args,0,x,y)
		pcb.drill[y][x]=0
		if args.verbose:
			print "; drilling @ {},{}".format(pcb.width+1-x,pcb.height+1-y)
		downdrill(args)
		up(args)
	gg(args,0,0,0)

def usage():
	print "Usage:"
	print "  "+sys.argv[0]+" filename"
	sys.exit(1)

def file_load(pcb):
	try:
		try:
			name=pcb.filename
	 		with open(name, "r") as f:
				loaded=json.load(f)
				for f in loaded:
					pcb.__dict__[f]=loaded[f]
	 		pcb.filename=name
	 	except TypeError:
#old format
	 		with open(pcb.filename, "r") as f:
				pcb.size,pcb.table,pcb.drill=json.load(f)
			pcb.height=len(pcb.table)-2
			pcb.width=len(pcb.table[0])-2
	except IOError:
		return False
	return True

if __name__ == '__main__':
	main()

